/*********************************************************************

   RW_BAR

   Driver basico PCI que permite operaciones de filesystem en todos
   los BAR disponibles en el dispositivo.

   sebfer@fing.edu.uy
   ciro@mondueri.com

   $ID$

 ********************************************************************
 Ver: 1.2
 - chequeo de offset y count para no salirse de la region de memoria
   asignada al bar correspondiente
 - mejora de estadisticas

 Ver: 1.1
 - vendor id y device id son parametros del modulo

 Ver: 1.0
 - asigna major number on the fly
 - minor number corresponde a la RAM visible para ese BAR

 ********************************************************************
 Notas: 
 - Usar load_rw_bar.sh para crear los nodos de device.


 ********************************************************************/

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/errno.h>
#include <linux/pci.h>
#include <linux/init.h>
#include <linux/poll.h>
#include <asm/system.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/uaccess.h>  /* copy_to_user y __copy_from_user */

#include <linux/proc_fs.h>
#include <asm/msr.h>

/* vendor id y device id core altera
 * deben estar definidos como constantes */
#define MY_DEVICE_ID 0xabba
#define MY_DEVICE_VENDOR 0x1172

/* debug code handing */
#undef PDEBUG
#ifdef DEBUG_RW_BAR
# define PDEBUG(fmt, args...) printk( "rw_bar: " fmt, ## args)
#else
# define PDEBUG(fmt, args...)
#endif
/* debug placeholder para deshabilitar facilmente una linea de debug */
#undef PDEBUGG
#define PDEBUGG(fmt, args...)

/* loops to usecs */
#define LOOPS2USECS(loops) (((double)loops*((double)1000000/((double)HZ*(double)current_cpu_data.loops_per_jiffy))))

/* bytes per sec */
#define KBYTESPERSEC(bytes, loops) ((((double)(bytes)/(double)loops)*(HZ*current_cpu_data.loops_per_jiffy))/((double)1024))

/* bar values para ioremap*/
unsigned long mem_base[6]; 
unsigned long mem_size[6];
void *mem_v_base[6];

/* parametros del modulo */
int rw_bar_major = 0;
int used_bars = 1;

MODULE_LICENSE("GPL");
MODULE_AUTHOR("sebfer@fing.edu.uy - ciro@mondueri.com");
MODULE_PARM(rw_bar_major, "i");
MODULE_PARM_DESC(rw_bar_major, "Major node to be assigned. Dynamic allocation if 0 or unspecified");
MODULE_PARM(used_bars, "i");
MODULE_PARM_DESC(used_bars, "Number of BAR registers to be mapped by the driver (MAX 6)");

#ifdef DRIVER_STATS
/* variables para estadisticas */
/* memcopy (de ram a tarjeta pci) */
unsigned long st_mc_w_total_bytes;
unsigned long st_mc_w_total_tsc_loops = 1;
unsigned long st_mc_r_total_bytes;
unsigned long st_mc_r_total_tsc_loops = 1;
/* para bars 1 a 3 */
unsigned long st_mc_w_bar1_bytes;
unsigned long st_mc_w_bar1_tsc_loops= 1;
unsigned long st_mc_r_bar1_bytes;
unsigned long st_mc_r_bar1_tsc_loops= 1;
unsigned long st_mc_w_bar2_bytes;
unsigned long st_mc_w_bar2_tsc_loops= 1;
unsigned long st_mc_r_bar2_bytes;
unsigned long st_mc_r_bar2_tsc_loops= 1;
unsigned long st_mc_w_bar3_bytes;
unsigned long st_mc_w_bar3_tsc_loops= 1;
unsigned long st_mc_r_bar3_bytes;
unsigned long st_mc_r_bar3_tsc_loops= 1;

/* ultima lectura y ultima escritura */
unsigned long st_mc_w_last_bytes;
unsigned long st_mc_w_last_tsc_start;
unsigned long st_mc_w_last_tsc_end;
unsigned long st_mc_r_last_bytes;
unsigned long st_mc_r_last_tsc_start;
unsigned long st_mc_r_last_tsc_end= 1;


#endif 

/* ***********************************************************************/
/* FILE OPERATIONS *******************************************************/
/* ***********************************************************************/

/* 
 * fop_read
 */
static ssize_t fop_read (struct file *filp, char *buf, size_t count, loff_t *f_pos)
{
  PDEBUG("fop_read: entra\n");

  int retval;
  unsigned char *kbuf, *ptr;
  void *add;
  int bar_to_read = MINOR(filp->f_dentry->d_inode->i_rdev);


  /*
   * no permito leer por fuera del espacio asignado
   */
  if ( *f_pos >= mem_size[bar_to_read] ) return 0;

  if ( (*f_pos + count) > mem_size[bar_to_read] ) 
    count = mem_size[bar_to_read] - 1 - *f_pos;

  if (count < 0) return 0;


  PDEBUG("fop_read: kmalloc antes\n");
  PDEBUG("fop_read: kmalloc count: %i\n", count);

  kbuf = kmalloc(count, GFP_KERNEL);
  if (!kbuf) return -ENOMEM;
  ptr=kbuf;
  retval=count;


  /*
   * Switch over to our remapped address space.
   */
  add = mem_v_base[bar_to_read] + *f_pos;

  PDEBUG("fop_read: memcopy antes\n");
  PDEBUG("fop_read: memcopy v_addres: %x\n", (int)add);
  PDEBUG("fop_read: memcopy_fromio ptr %x, add %x, count %i\n", (int)ptr, (int)add, count);

#ifdef DRIVER_STATS
  rdtscl(st_mc_r_last_tsc_start);
  /* NO PONER NADA ENTRE ESTO Y MEMCPY_FROMIO -> AFECTA ESTADISTICAS */
#endif
  memcpy_fromio(ptr, add, count);
#ifdef DRIVER_STATS
  rdtscl(st_mc_r_last_tsc_end);
  st_mc_r_total_tsc_loops += st_mc_r_last_tsc_end - st_mc_r_last_tsc_start;
  st_mc_r_total_bytes += count;
  st_mc_r_last_bytes = count;
  switch (bar_to_read) {
  case 1 : 
    st_mc_r_bar1_tsc_loops += st_mc_r_last_tsc_end - st_mc_r_last_tsc_start;
    st_mc_r_bar1_bytes += count;
    break;
  case 2 : 
    st_mc_r_bar1_tsc_loops += st_mc_r_last_tsc_end - st_mc_r_last_tsc_start;
    st_mc_r_bar1_bytes += count;
    break;
  case 3 : 
    st_mc_r_bar1_tsc_loops += st_mc_r_last_tsc_end - st_mc_r_last_tsc_start;
    st_mc_r_bar1_bytes += count;
    break;    
  }

#endif

  PDEBUG("fop_read: copy_to_user antes\n");
  PDEBUG("fop_read: copy_to_user: kbuf: %x\n", (int)kbuf);

  if (retval > 0)
    copy_to_user(buf, kbuf, retval);
  kfree(kbuf);
  *f_pos += retval;
  return retval;
}


/*
 * fop_write
 */
static ssize_t fop_write (struct file *filp, const char *buf, size_t count, loff_t *f_pos)
{
  PDEBUG("fop_write: entra\n");
  PDEBUG("fop_write: count: %i\n", count);

  int retval;
  unsigned char *kbuf, *ptr;
  void *add;
  int bar_to_read = MINOR(filp->f_dentry->d_inode->i_rdev);

  /*
   * Writing is dangerous.
   * Allow root-only, independently of device permissions
   */
  if (!capable (CAP_SYS_RAWIO)) return -EPERM;

  /*
   * no permito leer por fuera del espacio asignado
   */
  if ( *f_pos >= mem_size[bar_to_read] ) return 0;

  if ( (*f_pos + count) > mem_size[bar_to_read] ) 
    count = mem_size[bar_to_read] - 1 - *f_pos;

  if (count < 0) return 0;



  PDEBUG("fop_write: kmalloc antes\n");

  kbuf = kmalloc(count, GFP_KERNEL);
  if (!kbuf) return -ENOMEM;
  ptr=kbuf;
  retval=count;

  PDEBUG("fop_write: copy_from_user antes\n");
  PDEBUG("fop_write: copy_from_user: kbuf: %x\n", (int)kbuf);

  __copy_from_user(kbuf, buf, count);
  ptr=kbuf;

  /*
   * Switch over to our remapped address space.
   */
  add = mem_v_base[bar_to_read] + *f_pos;

  PDEBUG("fop_write: memcpy_toio antes\n");
  PDEBUG("fop_write: memcopy v_addres: %x\n", (int)add);
  PDEBUG("fop_write: memcopy_toio add %x, ptr %x, count %i\n", (int)add, (int)ptr, count);
#ifdef DRIVER_STATS
  rdtscl(st_mc_w_last_tsc_start);
  /* NO PONER NADA ENTRE ESTO Y MEMCPY_TOIO -> AFECTA ESTADISTICAS */
#endif
  memcpy_toio(add, ptr, count);
#ifdef DRIVER_STATS
  rdtscl(st_mc_w_last_tsc_end);
  st_mc_w_total_tsc_loops += st_mc_w_last_tsc_end - st_mc_w_last_tsc_start;
  st_mc_w_total_bytes += count;
  st_mc_w_last_bytes = count;
  switch (bar_to_read) {
  case 1 : 
    st_mc_w_bar1_tsc_loops += st_mc_w_last_tsc_end - st_mc_w_last_tsc_start;
    st_mc_w_bar1_bytes += count;
    break;
  case 2 : 
    st_mc_w_bar1_tsc_loops += st_mc_w_last_tsc_end - st_mc_w_last_tsc_start;
    st_mc_w_bar1_bytes += count;
    break;
  case 3 : 
    st_mc_w_bar1_tsc_loops += st_mc_w_last_tsc_end - st_mc_w_last_tsc_start;
    st_mc_w_bar1_bytes += count;
    break;    
  }
#endif

  PDEBUG("fop_write: kfree antes\n");

  *f_pos += retval;
  kfree(kbuf);
  return retval;
}


/*
 * fop_poll
 */
static unsigned int fop_poll(struct file *filp, poll_table *wait)
{
  return (POLLIN | POLLRDNORM | POLLOUT | POLLWRNORM);
}


/*
 * fop_open
 */
static int fop_open (struct inode *inode, struct file *filp)
{
  PDEBUG("fop_open: entra\n");
  // MOD_INC_USE_COUNT;
  return 0;
}


/*
 * fop_release
 */
static int fop_release (struct inode *inode, struct file *filp)
{
  PDEBUG("fop_release: entra\n");
  // MOD_DEC_USE_COUNT;
  return (0);
}


/* 
 * file operations
 */
static struct file_operations rw_bar_fops = {
  read: fop_read,
  write: fop_write,
  poll: fop_poll,
  open: fop_open,
  release: fop_release,
};


/* ***********************************************************************/
/* PCI DRIVER FUNCTIONS **************************************************/
/* ***********************************************************************/

/*
 * Cleanup del driver PCI (REMOVE)
 */
static void __devexit rw_bar_remove_one (struct pci_dev *pdev)
{
  PDEBUG("cleanup module sale\n");

};


/*
 * Inicializacion del driver PCI (PROBE).
 */
static int __devinit rw_bar_init_one (struct pci_dev *pdev,
				      const struct pci_device_id *ent)
{

  // hay 6 bars nada mas
  if (used_bars > 5) {
    used_bars = 6;
  }

  // para cada bar leer los valores
  int i;
  for (i = 0; i<used_bars; i++) {
    mem_base[i] = pci_resource_start(pdev, i);
    mem_size[i] = pci_resource_len(pdev, i);
    mem_v_base[i] = 0;

    PDEBUG("__devinit: mem_size[%i]: %x \n", i, (int)mem_size[i]);
    PDEBUG("__devinit: mem_base[%i]: %x \n", i, (int)mem_base[i]);
    PDEBUG("__devinit: mem_v_size[%i]: %x \n", i, (int)mem_v_base[i]);


    if ( (mem_base[i] <= 0) || (mem_size[i] <= 0) ) {
      if (i == 0 ) { 
	return -ENODEV;
      }
      
      continue;
    }
  }

  return 0;
}


/*
 * dispositivos en los que se usa este driver
 */
static struct pci_device_id rw_bar_tbl[] __devinitdata = {
  { MY_DEVICE_VENDOR, MY_DEVICE_ID, PCI_ANY_ID, PCI_ANY_ID, 0, 0, 0 },
};

MODULE_DEVICE_TABLE(pci, rw_bar_tbl);


/*
 * estructura del driver PCI
 */

static struct pci_driver rw_bar_driver = {
  name:		"rw_bar",
  probe:		rw_bar_init_one,
  remove:		rw_bar_remove_one,
  id_table:	rw_bar_tbl,
};



/* ***********************************************************************/
/* MODULE FUNCTIONS ******************************************************/
/* ***********************************************************************/

/*
 * read_procmem
 */
int rw_bar_read_procmem(char *buf, char **start, off_t offset, 
			int count, int *eof, void *data)
{
  int len = 0;
  len += sprintf(buf+len, "\n");
  len += sprintf(buf+len, "--------- IEEPCI Stats - HZ : %4i - loops_per_jiffy: %8li -------\n", HZ, current_cpu_data.loops_per_jiffy);
  len += sprintf(buf+len, "\n");
  len += sprintf(buf+len, "--------- W R I T E  --  M E M C O P Y -------------------------------\n");
  len += sprintf(buf+len, "          |   KB/sec     |     bytes    | microseconds | tsc loops    \n");
  len += sprintf(buf+len, "----------+--------------+--------------+--------------+--------------\n");
  len += sprintf(buf+len, "TOTAL     |%14i|%14li|%14i|%14li\n",(int)KBYTESPERSEC(st_mc_w_total_bytes, st_mc_w_total_tsc_loops), st_mc_w_total_bytes, (int)LOOPS2USECS(st_mc_w_total_tsc_loops), st_mc_w_total_tsc_loops);
  len += sprintf(buf+len, "BAR1 TOTAL|%14i|%14li|%14i|%14li\n",(int)KBYTESPERSEC(st_mc_w_bar1_bytes, st_mc_w_bar1_tsc_loops), st_mc_w_bar1_bytes, (int)LOOPS2USECS(st_mc_w_bar1_tsc_loops), st_mc_w_bar1_tsc_loops);
  len += sprintf(buf+len, "BAR2 TOTAL|%14i|%14li|%14i|%14li\n",(int)KBYTESPERSEC(st_mc_w_bar2_bytes, st_mc_w_bar2_tsc_loops), st_mc_w_bar2_bytes, (int)LOOPS2USECS(st_mc_w_bar2_tsc_loops), st_mc_w_bar2_tsc_loops);
  len += sprintf(buf+len, "BAR3 TOTAL|%14i|%14li|%14i|%14li\n",(int)KBYTESPERSEC(st_mc_w_bar3_bytes, st_mc_w_bar3_tsc_loops), st_mc_w_bar3_bytes, (int)LOOPS2USECS(st_mc_w_bar3_tsc_loops), st_mc_w_bar3_tsc_loops);
  len += sprintf(buf+len, "LAST WRITE|%14i|%14li|%14i|%14li\n",(int)KBYTESPERSEC(st_mc_w_last_bytes, (st_mc_w_last_tsc_end-st_mc_w_last_tsc_start)), st_mc_w_last_bytes, (int)LOOPS2USECS((st_mc_w_last_tsc_end-st_mc_w_last_tsc_start)), st_mc_w_last_tsc_end-st_mc_w_last_tsc_start);
  len += sprintf(buf+len, "--------- R E A D  --  M E M C O P Y ---------------------------------\n");
  len += sprintf(buf+len, "          |   KB/sec     |     bytes    | microseconds | tsc loops    \n");
  len += sprintf(buf+len, "----------+--------------+--------------+--------------+--------------\n");
  len += sprintf(buf+len, "TOTAL     |%14i|%14li|%14i|%14li\n",(int)KBYTESPERSEC(st_mc_r_total_bytes, st_mc_r_total_tsc_loops), st_mc_r_total_bytes, (int)LOOPS2USECS(st_mc_r_total_tsc_loops), st_mc_r_total_tsc_loops);
  len += sprintf(buf+len, "BAR1 TOTAL|%14i|%14li|%14i|%14li\n",(int)KBYTESPERSEC(st_mc_r_bar1_bytes, st_mc_r_bar1_tsc_loops), st_mc_r_bar1_bytes, (int)LOOPS2USECS(st_mc_r_bar1_tsc_loops), st_mc_r_bar1_tsc_loops);
  len += sprintf(buf+len, "BAR2 TOTAL|%14i|%14li|%14i|%14li\n",(int)KBYTESPERSEC(st_mc_r_bar2_bytes, st_mc_r_bar2_tsc_loops), st_mc_r_bar2_bytes, (int)LOOPS2USECS(st_mc_r_bar2_tsc_loops), st_mc_r_bar2_tsc_loops);
  len += sprintf(buf+len, "BAR3 TOTAL|%14i|%14li|%14i|%14li\n",(int)KBYTESPERSEC(st_mc_r_bar3_bytes, st_mc_r_bar3_tsc_loops), st_mc_r_bar3_bytes, (int)LOOPS2USECS(st_mc_r_bar3_tsc_loops), st_mc_r_bar3_tsc_loops);
  len += sprintf(buf+len, "LAST READ |%14i|%14li|%14i|%14li\n",(int)KBYTESPERSEC(st_mc_r_last_bytes, (st_mc_r_last_tsc_end-st_mc_r_last_tsc_start)), st_mc_r_last_bytes, (int)LOOPS2USECS((st_mc_r_last_tsc_end - st_mc_r_last_tsc_start)), st_mc_r_last_tsc_end-st_mc_r_last_tsc_start);

  *eof = 1;
  return len;

}
			


/*********************************************************************
 * init_module
 ********************************************************************/
int __init rw_bar_init(void)
{
  int result;
  int i;
  
  PDEBUG("rw_bar_init: antes pci_module_init\n");
  PDEBUG("rw_bar_init: used_bars: %i \n", used_bars);
  PDEBUG("rw_bar_init: vendor: %x \n", (int)rw_bar_tbl->vendor);
  PDEBUG("rw_bar_init: device: %x \n", (int)rw_bar_tbl->device);


  /* 
   * cargo el driver PCI.
   * pci_module_init detecta si el dispositivo especificado en
   * rw_bar_driver esta instalado en el sistema y en dicho caso
   *  ejecuta rw_bar_init_one 
   */
  if ( pci_module_init (&rw_bar_driver) != 0){
    PDEBUG("rw_bar_init: pci_module_init != 0\n");
    return -ENODEV;
  }


#ifdef DEBUG_RW_BAR
  for (i=0; i<used_bars; i++) {
    PDEBUG("rw_bar_init: mem_size[%i]: %x \n", i, (int)mem_size[i]);
    PDEBUG("rw_bar_inite: mem_base[%i]: %x \n", i, (int)mem_base[i]);
  }
#endif


  /*
   * Reservo la region fisica de memoria correspondiente a cada BAR
   */
  for (i = 0; i<used_bars; i++) {
    if (check_mem_region(mem_base[i], mem_size[i])) {
      printk("drivername: memory already in use\n");
      return -EBUSY;
    }
    request_mem_region(mem_base[i], mem_size[i], "rw_bar");

    /*
     * Mapear memoria fisica dispositivo a memoria virtual
     */
    mem_v_base[i] = ioremap(mem_base[i], mem_size[i]);
  }


  /*
   * registro del character device (fops)
   */
  result = register_chrdev(rw_bar_major, "rw_bar", &rw_bar_fops);
  if (result < 0) {
    printk(KERN_INFO "rw_bar: can't get major number\n");
    return result;
  }
  if (rw_bar_major == 0)
    rw_bar_major = result; /* dynamic */


  /*
   * registro de entrada proc/drivers/iiepci
   */
  create_proc_read_entry("driver/iiepci",
			 0 /* default mode */,
			 NULL /*parent dir */,
			 rw_bar_read_procmem,
			 NULL /* client data */);

  return 0;
}




/*********************************************************************
 * cleanup_module
 ********************************************************************/
void __exit rw_bar_cleanup(void)
{

  /* unregister proc/drivers/ieepci entry */
  remove_proc_entry("driver/iiepci", NULL /* parent dir */);

  /*
   * unregister memory
   */
  int i;
  for (i = 0; i<used_bars; i++) {
    iounmap((void *)mem_v_base[i]);
    release_mem_region(mem_base[i], mem_size[i]);  
  }

  /*
   * unregister el character device
   */
  unregister_chrdev(rw_bar_major, "rw_bar");

  /*
   * unregister PCI driver
   */
  pci_unregister_driver (&rw_bar_driver);

}


/*********************************************************************
 * Hooks de inicializacion y salida del MODULO.
 ********************************************************************/
module_init(rw_bar_init);
module_exit(rw_bar_cleanup);


